(ns microtables-frontend.evaluation.impl
  (:require
   ["mathjs" :as mathjs]
   [clojure.set :refer [intersection]]
   [clojure.string :as string]
   [microtables-frontend.utils.data :as data-utils]))

; the references in the data are a set of disconnected, doubly-linked trees
;TODO: rather than denotify all, then re-notify all, maybe use a diff? maybe on small scales it's not worth it?
(defn denotify-references
  "Remove references in all cells formerly referenced by this cell"
  [data origin refs]
  (if (empty? refs)
    data
    (let [target (first refs)
          de-notified (update-in data [(:col target) (:row target) :inbound] (partial filter #(not= % origin)))]
      (recur de-notified origin (rest refs)))))

(defn notify-references
  "Update references in all cells referenced by this cell"
  [data origin refs]
  (if (empty? refs)
    data
    (let [target (first refs)
          notified (update-in data [(:col target) (:row target) :inbound] conj origin)]
      (recur notified origin (rest refs)))))

(defn- remove-valueless-range-elements
  "Remove nil values specifically from ranges (to solve issues with some functions like average)."
  [variables var-list]
  (let [l (string/split (string/replace (first var-list) #"[()]" "") #",")
        has-values (filter #(not (nil? (variables %))) l)]
    (str "(" (string/join "," has-values) ")")))

(defn preprocess-expression
  "Handle range cases, rename certain functions (to work with math.js), prepare expression and variables for processing."
  [expression variables]
  (let [renamed-expression (string/replace expression #"\baverage\(" "mean(")
        new-expression (string/replace renamed-expression
                                       #"\(([A-Z]+[0-9]+,)*[A-Z]+[0-9]+\)"
                                       (partial remove-valueless-range-elements variables))
        new-variables (reduce-kv #(assoc %1 %2 (if (nil? %3) "0" %3)) {} variables)]
    (println "PREPROCESS" {:expression new-expression :variables new-variables})
    {:expression new-expression
     :variables new-variables}))

;TODO: memoize dynamically? probably not worth memoizing directly, and could take up too much memory over time
;      https://stackoverflow.com/a/13123571/8172807
(defn find-cycle
  "Accepts the data and a datum, and peforms a depth-first search to find reference cycles, following back-references."
  ([data c r] (find-cycle data c r #{}))
  ([data c r ancest]
   (let [datum (get-in data [c r])
         current {:col c :row r}
         this-and-above (conj ancest current)
         inbound (:inbound datum)
         found-repeat (not (empty? (intersection this-and-above (set inbound))))]
     (if found-repeat
       :cycle-error
       (some #(find-cycle data (:col %) (:row %) this-and-above) inbound)))))

(def evaluate-expression
  "Convert (via mathjs) an expression string to a final answer (also a string).
  A map of variables must also be provided. If there is an error, it will return :calc-error."
  (memoize (fn [expression variables]
             (let [range-replaced (data-utils/replace-ranges-in-expression expression)
                   {ready-expression :expression
                    ready-variables :variables} (preprocess-expression range-replaced variables)]
               (try
                 (.evaluate mathjs ready-expression (clj->js ready-variables))
                 (catch js/Error e
                   (println "mathjs evaluation error" (.-message e) e)
                   :calc-error))))))

(defn gather-variables-and-evaluate-cell
  "Assumes that all the cell's immediate references have been resolved.
  Collects the final values from them, then evaluates the current cell's expression.
  Returns the new data map."
  [data c r]
  (let [datum (dissoc (dissoc (get-in data [c r]) :dirty) :display) ; get rid of the dirty flag right away (it must be included with the returned data to have effect)
        refs (:refs datum)
        value (:value datum)
        formula (data-utils/formula? value)
        resolved-refs (map #(merge % (get-in data [(:col %) (:row %)])) refs)
        evaluated-refs (map #(if (data-utils/formula? (:value %)) (:display %) (:value %)) resolved-refs)
        invalid-refs (some nil? resolved-refs)
        dirty-refs (some :dirty resolved-refs)
        error-refs (some #(= (:display %) :error) resolved-refs)
        ;unevaluated-refs (some nil? evaluated-refs)
        cycle-refs (some #(= (:display %) :cycle-error) resolved-refs)
        disqualified? (or invalid-refs dirty-refs error-refs)]
    (cond
      (not formula) (assoc-in data [c r] datum)     ; if it's not a formula, then return as is (with the dirty flag removed)
      cycle-refs (-> data                           ; if one of its references has a reference cycle, then this one is "poisoned" as well
                     (assoc-in [c r] datum)
                     (assoc-in [c r :display] :cycle-error))
      ;unevaluated-refs (assoc-in data [c r :display] :insufficient-data) ; do not un-mark as "dirty", since it has not been evaluated yet
      disqualified? (-> data                        ; some other error is present
                        (assoc-in [c r] datum)
                        (assoc-in [c r :display] :error))
      (empty? refs) (-> data
                        (assoc-in [c r] datum)
                        (assoc-in [c r :display] (evaluate-expression (subs value 1) {})))
      :else (let [variables (zipmap (map #(str (:col %) (:row %)) refs) evaluated-refs)
                  evaluated-value (evaluate-expression (subs value 1) variables)
                  new-datum (assoc datum :display evaluated-value)]
              (assoc-in data [c r] new-datum)))))
